import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"
import pytest
import datetime as dt

from unittest.mock import MagicMock, patch

from sorunlib.commands import wait_until


def mkts(offset):
    """Make timestamp.

    Args:
        offset (int): Offset from current time in seconds.

    Returns:
        str: ISO formatted timestamp 'offset' seconds from now, i.e.
            '2023-04-22T00:59:56.264293+00:00'

    Examples:
        An example called at '2023-04-24T21:14:27.790440+00:00':

        >>> mkts(1)
        '2023-04-24T21:14:28.790440+00:00'

    """
    now = dt.datetime.now(dt.timezone.utc)
    delta = dt.timedelta(seconds=offset)
    ts = now + delta
    return ts.isoformat()


# patch out time.sleep so we don't actually sleep during testing
@patch('sorunlib.commands.time.sleep', MagicMock())
@pytest.mark.parametrize("timestamp,tolerance", [
    # timestamp in past, not high enough tolerance
    (mkts(-10), 5),
    # timestamp in future, but past tolerance timestamp
    (mkts(1), mkts(-1))])
def test_wait_until_past_tolerance(timestamp, tolerance):
    with pytest.raises(ValueError):
        wait_until(timestamp, tolerance)


@patch('sorunlib.commands.time.sleep', MagicMock())
def test_wait_until_unsupported_tolerance():
    with pytest.raises(ValueError):
        wait_until(mkts(1), tolerance=[1, 2])


def test_wait_until_unsupported_tz():
    with pytest.raises(ValueError):
        tz = dt.timezone(offset=dt.timedelta(hours=5))
        t = dt.datetime.now(tz).isoformat()  # i.e. '2023-04-22T00:59:56.264293+05:00'
        wait_until(t)


@patch('sorunlib.commands.time.sleep', MagicMock())
@pytest.mark.parametrize("timestamp,tolerance", [
    # timestamps in future, future or no tolerance
    (mkts(1), None),
    (mkts(1), 5),
    (mkts(1), mkts(10)),
    # timestamps in past, high enough or no tolerance
    (mkts(-1), None),
    (mkts(-1), 5),
    (mkts(-1), mkts(10)),
    # test mix of tz aware timestamp, naive tolerance timestamp
    (mkts(0), mkts(10)[:-6]),
    (mkts(0)[:-6], mkts(10)),
    # testing TZ detection w/past timestamps, no tolerance
    ("2020-01-01T00:00:00", None),
    ("2020-01-01T00:00:00+00:00", None)])
def test_wait_until(timestamp, tolerance):
    wait_until(timestamp, tolerance)



import os
import ocs
import pytest
import datetime as dt

from unittest.mock import MagicMock, patch
from ocs.ocs_client import OCSReply

from sorunlib._internal import check_response, check_running, check_started, monitor_process

from util import create_session as create_unencoded_session

os.environ["OCS_CONFIG_DIR"] = "./test_util/"


def create_session(*args, **kwargs):
    session = create_unencoded_session(*args, **kwargs)
    return session.encoded()


# Very partial mock as we only need the instance_id to exist.
class MockClient:
    def __init__(self):
        self.instance_id = 'test-id'
        self.test_op = MagicMock()


# Attempt to recreate https://github.com/simonsobs/sorunlib/issues/174
malformed_session = create_session('test', success=False)
malformed_session.pop('op_name')

invalid_responses = [(MockClient(), OCSReply(ocs.TIMEOUT,
                                             'msg',
                                             create_session('test', success=True))),
                     (MockClient(), OCSReply(ocs.ERROR,
                                             'msg',
                                             create_session('test', success=False))),
                     (MockClient(), OCSReply(ocs.ERROR,
                                             'msg',
                                             malformed_session))]

valid_responses = [
    (MockClient(), OCSReply(ocs.OK, 'msg', create_session('test', success=True)))]

invalid_running_responses = [
    (MockClient(), OCSReply(ocs.OK, 'msg', create_session('test', status='done')))]

running_responses = [
    (MockClient(), OCSReply(ocs.OK, 'msg', create_session('test', status='running')))]

invalid_starting_responses = [
    (MockClient(), OCSReply(ocs.OK, 'msg', create_session('test', status='starting')))]


@pytest.mark.parametrize("client,response", invalid_responses)
def test_check_response_raises(client, response):
    with pytest.raises(RuntimeError):
        check_response(client, response)


@pytest.mark.parametrize("client,response", valid_responses)
def test_check_response(client, response):
    check_response(client, response)


@pytest.mark.parametrize("client,response", invalid_running_responses)
def test_check_running_raises(client, response):
    with pytest.raises(RuntimeError):
        check_running(client, response)


@pytest.mark.parametrize("client,response", running_responses)
def test_check_running(client, response):
    check_running(client, response)


@patch('sorunlib._internal.time.sleep', MagicMock())
@pytest.mark.parametrize("client,response", invalid_starting_responses)
def test_check_started_raises_timeout(client, response):
    with pytest.raises(RuntimeError):
        check_started(client, response)


@pytest.mark.parametrize("client,response", invalid_running_responses)
def test_check_started_raises(client, response):
    with pytest.raises(RuntimeError):
        check_started(client, response, timeout=1)


@pytest.mark.parametrize("client,response", running_responses)
def test_check_started(client, response):
    check_started(client, response)


@pytest.mark.parametrize("client,response", running_responses)
def test_monitor_process(client, response):
    # Prepare client for status check
    client.test = MagicMock()
    client.test.status = MagicMock(return_value=response)
    # Durations here affect test runtime, keep them short
    stop_time = (dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=0.01)).isoformat()
    monitor_process(client, 'test', stop_time, check_interval=0.005)


@pytest.mark.parametrize("client,response", invalid_running_responses)
def test_monitor_process_invalid(client, response):
    # Prepare client for status check
    client.test = MagicMock()
    client.test.status = MagicMock(return_value=response)
    stop_time = (dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=0.01)).isoformat()
    with pytest.raises(RuntimeError):
        monitor_process(client, 'test', stop_time, check_interval=1)



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"
import datetime as dt

import pytest

from ocs.ocs_client import OCSReply
from sorunlib import acu

from util import create_patch_clients


patch_clients_satp = create_patch_clients('satp')
patch_clients_lat = create_patch_clients('ccat')


def test_move_to(patch_clients_satp):
    acu.move_to(180, 60)
    acu.run.CLIENTS['acu'].go_to.assert_called_with(az=180, el=60)


def test_move_to_failed(patch_clients_satp):
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'go_to'})
    acu.run.CLIENTS['acu'].go_to.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        acu.move_to(180, 90)


def test_move_to_target_before_start(patch_clients_satp):
    start = dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=10)
    end = start + dt.timedelta(seconds=3600)
    acu.move_to_target(300, 50, start.isoformat(), end.isoformat(), -0.005)
    acu.run.CLIENTS['acu'].go_to.assert_called_with(az=300, el=50)


def test_move_to_target_within_range(patch_clients_satp):
    start = dt.datetime.now(dt.timezone.utc)
    end = start + dt.timedelta(seconds=3600)
    acu.move_to_target(300, 50, start.isoformat(), end.isoformat(), -0.005)
    acu.run.CLIENTS['acu'].go_to.assert_called_once()


def test_move_to_target_after_stop(patch_clients_satp):
    start = dt.datetime.now(dt.timezone.utc) - dt.timedelta(seconds=3600)
    end = start + dt.timedelta(seconds=3590)
    acu.move_to_target(300, 50, start.isoformat(), end.isoformat(), -0.005)
    acu.run.CLIENTS['acu'].go_to.assert_not_called()


def test_set_boresight(patch_clients_satp):
    acu.set_boresight(20)
    acu.run.CLIENTS['acu'].set_boresight.assert_called_with(target=20)


def test_set_boresight_lat(patch_clients_lat):
    with pytest.raises(RuntimeError):
        acu.set_boresight(20)


def test_set_scan_params(patch_clients_satp):
    acu.set_scan_params(az_speed=2, az_accel=2, reset=True)
    acu.run.CLIENTS['acu'].set_scan_params.assert_called_with(
        az_speed=2,
        az_accel=2,
        reset=True)



import sorunlib as run
from sorunlib._internal import check_response


def _get_direction():
    """Get the rotational direction ('cw' or 'ccw') of the HWP. The direction
    is determined in part by the configuration of the HWP supervisor agent. For
    details see the `HWP Supervisor agent docs <docs_>`_.

    * 'cw' is clockwise as seen from the sky to window.
    * 'ccw' is counter-clockwise as seen from the sky to window.

    Returns:
        str: The direction of the HWP, either 'cw' or 'ccw'.

    Raises:
        RuntimeError: If the direction is not either 'cw' or 'ccw'.

    .. _docs: https://socs.readthedocs.io/en/main/agents/hwp_supervisor_agent.html

    """
    hwp = run.CLIENTS['hwp']
    resp = hwp.monitor.status()
    direction = resp.session['data']['hwp_state']['direction']

    if direction not in ['cw', 'ccw']:
        raise RuntimeError("The HWP direction is unknown. Aborting...")

    return direction


# Public API
def set_freq(freq):
    """Set the rotational frequency of the HWP.

    Args:
        freq (float): Target frequency to rotate the HWP in Hz. This is a
            *signed float*, the meaning of which depends on the OCS site
            configuration. For details see the `documentation for the HWP
            Supervisor Agent <docs_>`_.

    .. _docs: https://socs.readthedocs.io/en/main/agents/hwp_supervisor_agent.html

    """
    hwp = run.CLIENTS['hwp']
    resp = hwp.pid_to_freq(target_freq=freq)
    check_response(hwp, resp)


def stop(active=True, brake_voltage=None):
    """Stop the HWP.

    Args:
        active (bool, optional): If True, actively try to stop the HWP by
            applying the brake. If False, simply turn off the PMX power and wait
            for it to spin down on its own. Defaults to True.
        brake_voltage (float, optional): Voltage used when actively stopping
            the HWP. Only considered when active is True.

    """
    hwp = run.CLIENTS['hwp']

    print('Stopping HWP and waiting for it to spin down.')
    if active:
        if brake_voltage is None:
            resp = hwp.brake()
        else:
            resp = hwp.brake(brake_voltage=brake_voltage)
        check_response(hwp, resp)
    else:
        resp = hwp.pmx_off(wait_stop=True)
        check_response(hwp, resp)



from . import acu, hwp, seq, smurf, wiregrid

from .commands import wait_until
from .util import create_clients

CLIENTS = None


def initialize(test_mode=False):
    """Initialize global clients list.

    Args:
        test_mode (bool): Operate in 'test mode'. Use this to find Agents that
            are meant to stand in for real agents while testing, i.e.
            SmurfFileEmulators instead of PysmurfControllers.
    """
    global CLIENTS
    CLIENTS = create_clients(test_mode=test_mode)


__all__ = ["acu",
           "hwp",
           "seq",
           "smurf",
           "wiregrid",
           "wait_until",
           "initialize"]

from . import _version
__version__ = _version.get_versions()['version']



import os
import pytest

from ocs.client_http import ControlClientError
from unittest.mock import MagicMock, patch

from sorunlib import util
from sorunlib.util import CrossbarConnectionError

os.environ["OCS_CONFIG_DIR"] = "./test_util/"
os.environ["SORUNLIB_CONFIG"] = "./data/example_config.yaml"


# I think this could be generalized with the Mock 'spec' argument, building an
# object that mocks a given Agent's API, but I'm not totally sure we need that
# yet, or how to make it so a different spec is loaded depending on which agent
# the client is for.
def create_mock_ocsclient(session):
    def mock_client(*args, **kwargs):
        client = MagicMock()
        client.main.status = MagicMock(return_value=(None, None, session))
        return client

    return mock_client


reg_session = {'session_id': 0,
               'op_name': 'main',
               'op_code': 3,
               'status': 'running',
               'success': None,
               'start_time': 1669919099.7585046,
               'end_time': None,
               'data': {
                   'observatory.smurf-file-emulator-5': {
                       'expired': False,
                       'time_expired': None,
                       'last_updated': 1669935108.8366735,
                       'op_codes': {
                           'uxm_setup': 1,
                           'uxm_relock': 1,
                           'take_iv': 1,
                           'take_bias_steps': 1,
                           'take_bgmap': 1,
                           'bias_dets': 1,
                           'take_noise': 1,
                           'stream': 1},
                       'agent_class': 'SmurfFileEmulator',
                       'agent_address': 'observatory.smurf-file-emulator-5'},
                   'observatory.smurf-file-emulator-7': {
                       'expired': False,
                       'time_expired': None,
                       'last_updated': 1669935108.989246,
                       'op_codes': {
                           'uxm_setup': 1,
                           'uxm_relock': 1,
                           'take_iv': 1,
                           'take_bias_steps': 1,
                           'take_bgmap': 1,
                           'bias_dets': 1,
                           'take_noise': 1,
                           'stream': 1},
                       'agent_class': 'SmurfFileEmulator',
                       'agent_address': 'observatory.smurf-file-emulator-7'},
                   'observatory.fake-data-1': {
                       'expired': True,
                       'time_expired': None,
                       'last_updated': 1669935108.989246,
                       'op_codes': {
                           'acq': 3,
                           'count': 3,
                           'set_heartbeat': 1,
                           'delay_task': 1},
                       'agent_class': 'FakeDataAgent',
                       'agent_address': 'observatory.fake-data-1'},
                   'observatory.hwp-supervisor': {
                       'expired': False,
                       'time_expired': None,
                       'last_updated': 1710168924.336035,
                       'op_codes': {
                           'monitor': 3,
                           'spin_control': 3,
                           'pid_to_freq': 1,
                           'set_const_voltage': 1,
                           'brake': 1,
                           'pmx_off': 1,
                           'abort_action': 7
                       },
                       'agent_class': 'HWPSupervisor',
                       'agent_address': 'observatory.hwp-supervisor'
                   },
                   'observatory.acu-sat1': {
                       'expired': False,
                       'time_expired': None,
                       'last_updated': 1669997169.469505,
                       'op_codes': {
                           'monitor': 3,
                           'broadcast': 3,
                           'generate_scan': 1,
                           'go_to': 1,
                           'constant_velocity_scan': 1,
                           'fromfile_scan': 1,
                           'set_boresight': 1,
                           'stop_and_clear': 1},
                       'agent_class': 'ACUAgent',
                       'agent_address': 'observatory.acu-sat1'}}}
mock_registry_client = create_mock_ocsclient(reg_session)

reg_session_single_smurf = {'session_id': 0,
                            'op_name': 'main',
                            'op_code': 3,
                            'status': 'running',
                            'success': None,
                            'start_time': 1669919099.7585046,
                            'end_time': None,
                            'data': {
                                'observatory.smurf-file-emulator-5': {
                                    'expired': False,
                                    'time_expired': None,
                                    'last_updated': 1669935108.8366735,
                                    'op_codes': {
                                        'uxm_setup': 1,
                                        'uxm_relock': 1,
                                        'take_iv': 1,
                                        'take_bias_steps': 1,
                                        'take_bgmap': 1,
                                        'bias_dets': 1,
                                        'take_noise': 1,
                                        'stream': 1},
                                    'agent_class': 'SmurfFileEmulator',
                                    'agent_address': 'observatory.smurf-file-emulator-5'}}}
mock_registry_client_single_smurf = create_mock_ocsclient(reg_session_single_smurf)


class NoAgentClient:
    def __init__(self, *args, **kwargs):
        raise ControlClientError("no callee registered for procedure fake.agent.op")


class UnexpectedErrorClient:
    def __init__(self, *args, **kwargs):
        raise ControlClientError("Server replied with code 500")


def test_load_site_config():
    cfg = util._load_site_config()
    assert 'localhost' in cfg.hosts
    assert 'ocs-docker' in cfg.hosts
    assert 'registry_address' in cfg.hub.data


def test_find_instances():
    instances = util._find_instances('PysmurfController')
    assert instances == ['pysmurf-controller']


def test_find_instances_host():
    instances = util._find_instances('PysmurfController', host='localhost')
    assert instances == []


@patch('sorunlib.util.OCSClient', mock_registry_client)
def test_find_active_instances():
    instances = util._find_active_instances('SmurfFileEmulator')
    assert 'smurf-file-emulator-5' in instances
    assert 'smurf-file-emulator-7' in instances


@patch('sorunlib.util.OCSClient', mock_registry_client)
def test_find_active_instances_expired():
    instances = util._find_active_instances('FakeDataAgent')
    assert 'fake-data-1' not in instances


def test__try_client_no_crossbar_connection():
    """This test assumes a crossbar server isn't running at
    http://localhost:8001. It tests that we raise an error when trying to
    connect to the registry to scan for agents if crossbar is unavailable."""
    with pytest.raises(CrossbarConnectionError):
        util._try_client('test-agent')


@patch('sorunlib.util.OCSClient', NoAgentClient)
def test__try_client_no_agent_connection():
    """This tests that we get None back when the agent is offline."""
    client = util._try_client('test-agent')
    assert client is None


@patch('sorunlib.util.OCSClient', UnexpectedErrorClient)
def test__try_client_other_error():
    with pytest.raises(ControlClientError):
        util._try_client('test-agent')


@patch('sorunlib.util.OCSClient', mock_registry_client)
def test_create_clients():
    clients = util.create_clients()
    assert 'acu' in clients
    assert 'smurf' in clients
    assert len(clients['smurf']) == 0  # since we're not in test_mode
    assert 'wiregrid' in clients


@patch('sorunlib.util.OCSClient', mock_registry_client)
def test_create_clients_test_mode():
    clients = util.create_clients(test_mode=True)
    assert 'acu' in clients
    assert 'smurf' in clients
    assert len(clients['smurf']) == 2
    assert 'wiregrid' in clients


@patch('sorunlib.util.OCSClient', mock_registry_client_single_smurf)
def test_create_clients_test_mode_single_smurf():
    clients = util.create_clients(test_mode=True)
    assert 'smurf' in clients
    assert len(clients['smurf']) == 1


@patch('sorunlib.util.OCSClient', mock_registry_client)
def test__create_clients_minimal_config():
    clients = util.create_clients(sorunlib_config='./data/minimal_config.yaml')
    assert 'acu' in clients
    assert 'smurf' in clients
    assert len(clients['smurf']) == 0  # since we're not in test_mode
    # Optional configs
    assert 'wiregrid' in clients
    for client in clients['wiregrid'].values():
        assert client is None



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"
os.environ["SORUNLIB_CONFIG"] = "./data/example_config.yaml"
import datetime as dt

import ocs
import pytest
from unittest.mock import MagicMock, patch
from ocs.ocs_client import OCSReply

import sorunlib
from sorunlib import seq

from util import create_patch_clients, create_session


patch_clients = create_patch_clients('satp')


@patch('sorunlib._internal.time.sleep', MagicMock())
def test_scan(patch_clients):
    # This affects test runtime duration keep it short
    target = dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=0.01)
    seq.scan(description='test', stop_time=target.isoformat(), width=20.)


@patch('sorunlib._internal.time.sleep', MagicMock())
def test_scan_passed_stop_time(patch_clients):
    # This affects test runtime duration keep it short
    target = dt.datetime.now(dt.timezone.utc) - dt.timedelta(seconds=10)
    seq.scan(description='test', stop_time=target.isoformat(), width=20.)
    seq.run.CLIENTS['acu'].generate_scan.start.assert_not_called()


@patch('sorunlib._internal.time.sleep', MagicMock())
def test_scan_passed_min_duration(patch_clients):
    # This affects test runtime duration keep it short
    target = dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=10)
    seq.scan(description='test', stop_time=target.isoformat(), width=20., min_duration=60)
    seq.run.CLIENTS['acu'].generate_scan.start.assert_not_called()


@patch('sorunlib.commands.time.sleep', MagicMock())
def test_scan_no_session(patch_clients):
    # Setup mock OCSReply without session object
    mock_reply = MagicMock()
    mock_reply.session = None
    sorunlib.CLIENTS['acu'].generate_scan.start = MagicMock(return_value=mock_reply)

    target = dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=1)
    with pytest.raises(Exception):
        seq.scan(description='test', stop_time=target.isoformat(), width=20.)


@patch('sorunlib.commands.time.sleep', MagicMock())
def test_scan_failed_to_start(patch_clients):
    # Setup mock OCSReply with failed status
    failed_session = create_session('generate_scan')
    failed_session.success = False
    failed_session.set_status('running')
    failed_session.add_message('A simulated error has occurred.')
    failed_session.set_status('done')

    mock_reply = OCSReply(ocs.OK, 'msg', failed_session.encoded())
    sorunlib.CLIENTS['acu'].generate_scan.start = MagicMock(return_value=mock_reply)
    sorunlib.CLIENTS['acu'].generate_scan.wait = MagicMock(return_value=mock_reply)

    # Example of failed reply this is trying to emulate.
    # OCSReply: OK : Operation "generate_scan" is currently not running (FAILED).
    #   generate_scan[session=15]; status=done with ERROR 116.5 s ago, took 68.8 s
    #   messages (4 of 4):
    #     1702337679.564 Status is now "starting".
    #     1702337679.565 Status is now "running".
    #     1702337748.356 Problems during scan
    #     1702337748.357 Status is now "done".
    #   other keys in .session: op_code, data
    print(mock_reply)

    target = dt.datetime.now(dt.timezone.utc) + dt.timedelta(seconds=10)
    with pytest.raises(RuntimeError):
        seq.scan(description='test', stop_time=target.isoformat(), width=20.)


@patch('sorunlib._internal.time.sleep', MagicMock())
def test_scan_failed_smurfs_on_shutdown(patch_clients):
    # Create failure on all three smurf agents
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'stream'})
    seq.run.CLIENTS['smurf'][0].stream.wait.side_effect = [mocked_response]
    seq.run.CLIENTS['smurf'][1].stream.wait.side_effect = [mocked_response]
    seq.run.CLIENTS['smurf'][2].stream.wait.side_effect = [mocked_response]

    seq._stop_scan()
    seq.run.CLIENTS['acu'].generate_scan.wait.assert_called()



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"

import pytest
from unittest.mock import MagicMock
import time

import ocs
from ocs.ocs_client import OCSReply
from sorunlib import hwp

from util import create_patch_clients, create_session

patch_clients_satp = create_patch_clients('satp')


def create_hwp_client(direction):
    """Create a HWP client with mock acq Process session.data.

    Args:
        direction (str): direction of the HWP. 'ccw' (counter-clockwise) or 'cw' (clockwise).

    """
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        'hwp_state': {
            'direction': direction,
        },
        'timestamp': time.time(),
    }
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.monitor.status = MagicMock(return_value=reply)

    return client


@pytest.mark.parametrize('direction', ['ccw', 'cw'])
def test__get_direction(patch_clients_satp, direction):
    hwp.run.CLIENTS['hwp'] = create_hwp_client(direction)
    ret = hwp._get_direction()
    if direction == 'ccw':
        assert ret == 'ccw'
    elif direction == 'cw':
        assert ret == 'cw'
    hwp.run.CLIENTS['hwp'].monitor.status.assert_called_once()


@pytest.mark.parametrize('direction', [None, ''])
def test__get_direction_invalid(patch_clients_satp, direction):
    hwp.run.CLIENTS['hwp'] = create_hwp_client(direction)
    with pytest.raises(RuntimeError) as e:
        hwp._get_direction()
    assert str(e.value) == "The HWP direction is unknown. Aborting..."
    hwp.run.CLIENTS['hwp'].monitor.status.assert_called_once()


@pytest.mark.parametrize("active", [True, False])
def test_stop(patch_clients_satp, active):
    hwp.stop(active=active)
    if active:
        hwp.run.CLIENTS['hwp'].brake.assert_called_with()
    else:
        hwp.run.CLIENTS['hwp'].pmx_off.assert_called()


def test_stop_brake_voltage(patch_clients_satp):
    VOLTAGE = 5.0
    hwp.stop(active=True, brake_voltage=VOLTAGE)
    hwp.run.CLIENTS['hwp'].brake.assert_called_with(brake_voltage=VOLTAGE)


def test_set_freq(patch_clients_satp):
    hwp.set_freq(freq=2.0)
    hwp.run.CLIENTS['hwp'].pid_to_freq.assert_called_with(target_freq=2.0)



import os
import time
os.environ["OCS_CONFIG_DIR"] = "./test_util/"
os.environ["SORUNLIB_CONFIG"] = "./data/example_config.yaml"

import pytest

from unittest.mock import MagicMock, patch, call

import ocs
from ocs.ocs_client import OCSReply
from sorunlib import wiregrid

from util import create_session, create_patch_clients, mocked_clients

patch_clients = create_patch_clients('satp')


def create_acu_client(az, el, boresight):
    """Create an ACU client with mock monitor Process session.data.

    Args:
        az (float): Azimuth position.
        el (float): Elevation position.
        boresight (float): Boresight position.

    """
    acu_client = MagicMock()
    session = create_session('monitor')
    session.data = {'StatusDetailed': {'Azimuth current position': az,
                                       'Elevation current position': el,
                                       'Boresight current position': boresight}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    acu_client.monitor.status = MagicMock(return_value=reply)

    return acu_client


def create_labjack_client():
    """Create a labjack client with mock acq Process session.data."""
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        "block_name": "sens",
        "data": {
            "AIN0C": 23.0,
            "AIN1C": 20.0,
            "AIN2C": 30.0,
        },
        "timestamp": time.time()
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


def create_actuator_client(motor, position):
    """Create an actuator client with mock acq Process session.data.

    Args:
        motor (int): Motor state, 0 is off, 1 is on.
        position (str): Position of the wiregrid, either 'inside' or 'outside'.

    """
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        'fields': {
            'motor': motor,
            'position': position
        },
        'timestamp': time.time(),
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


def create_kikusui_client():
    """Create a kikusui client with mock IV_acq Process session.data."""
    client = MagicMock()
    session = create_session('IV_acq')
    session.data = {
        'fields': {
            'kikusui': {
                'volt': 12.0,
                'curr': 3.0,
                'voltset': 12.0,
                'currset': 3.0,
                'status': 1,
            }
        },
        'timestamp': time.time(),
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.IV_acq.status = MagicMock(return_value=reply)

    return client


def create_encoder_client():
    """Create an encoder client with mock acq Process session.data."""
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        'fields': {
            'encoder_data': {'last_updated': time.time()},
        },
        'timestamp': time.time(),
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_insert():
    wiregrid.insert()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].insert.assert_called_with()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_insert_failed():
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'insert'})
    wiregrid.run.CLIENTS['wiregrid']['actuator'].insert.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        wiregrid.insert()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_eject():
    wiregrid.eject()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].eject.assert_called_with()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_eject_failed():
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'eject'})
    wiregrid.run.CLIENTS['wiregrid']['actuator'].eject.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        wiregrid.eject()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_rotate_continuous():
    wiregrid.rotate(continuous=True)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_v.assert_called_once_with(volt=12)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_c.assert_called_once_with(current=3.0)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_on.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_off.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_rotate_stepwise():
    wiregrid.rotate(continuous=False)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_v.assert_called_once_with(volt=12)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_c.assert_called_once_with(current=3.0)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].stepwise_rotation.assert_called_once_with(num_laps=1, stopped_time=10)


def test__verify_temp_response():
    session = create_session('test')
    session.data = {"data": {"TEST": 2}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    wiregrid._verify_temp_response(reply, 'TEST', 0)


def test__verify_temp_response_invalid():
    session = create_session('test')
    session.data = {"data": {"TEST": -10}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    with pytest.raises(RuntimeError):
        wiregrid._verify_temp_response(reply, 'TEST', 0)


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_temperature_sensors():
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid._check_temperature_sensors()
    wiregrid.run.CLIENTS['wiregrid']['labjack'].acq.status.assert_called_once()


def test__check_telescope_position(patch_clients):
    wiregrid._check_telescope_position()
    wiregrid.run.CLIENTS['acu'].monitor.status.assert_called_once()


@pytest.mark.parametrize('el,boresight', [(40, 0), (50, 10)])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_telescope_position_invalid(el, boresight):
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, el, boresight)
    with pytest.raises(RuntimeError):
        wiregrid._check_telescope_position()
    wiregrid.run.CLIENTS['acu'].monitor.status.assert_called_once()


@pytest.mark.parametrize('motor', [(0), (1)])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_motor_on(motor):
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=motor, position='inside')
    wiregrid._check_motor_on()
    if motor == 1:
        wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
        wiregrid.run.CLIENTS['wiregrid']['actuator'].motor_on.assert_not_called()
    elif motor == 0:
        wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
        wiregrid.run.CLIENTS['wiregrid']['actuator'].motor_on.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_motor_on_invalid_state():
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=3, position='inside')
    with pytest.raises(RuntimeError):
        wiregrid._check_motor_on()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_agents_online():
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='inside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid._check_agents_online()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].IV_acq.status.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['encoder'].acq.status.assert_called()
    wiregrid.run.CLIENTS['wiregrid']['labjack'].acq.status.assert_called_once()


@pytest.mark.parametrize('continuous, el, tag',
                         [(True, 50, 'wiregrid, wg_continuous'),
                          (False, 90, 'wiregrid, wg_stepwise, wg_el90')])
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_calibrate_stepwise(patch_clients, continuous, el, tag):
    # Setup all mock clients
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, el, 0)
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='inside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()

    wiregrid.calibrate(continuous=continuous)
    # All other internal functions tested separately, just make sure smurf
    # stream is run
    for client in wiregrid.run.CLIENTS['smurf']:
        client.stream.start.assert_called_with(
            tag=tag,
            subtype='cal',
            kwargs={
                "downsample_factor": None,
                "filter_disable": False},
        )
        client.stream.stop.assert_called()


def test__check_process_data_stale_data():
    with pytest.raises(RuntimeError):
        stale_time = time.time() - wiregrid.AGENT_TIMEDIFF_THRESHOLD
        wiregrid._check_process_data('test process', stale_time)


@pytest.mark.parametrize('position', [('inside'), ('outside')])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_wiregrid_position(position):
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(1, position)
    return_position = wiregrid._check_wiregrid_position()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
    assert return_position == position


@pytest.mark.parametrize('position', [('unknown'), ('')])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_wiregrid_position_invalid(position):
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(1, position)
    with pytest.raises(RuntimeError):
        wiregrid._check_wiregrid_position()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_cw():
    # Setup all mock clients
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 50, 0)
    wiregrid.run.hwp._get_direction = MagicMock(return_value='cw')  # cw
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='outside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid.run.wiregrid.rotate = MagicMock()

    wiregrid.time_constant(num_repeats=1)

    # just make sure bias_steps and streams because other functions are already
    # tested separately.
    expected_calls_of_bias_steps = [
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_cw'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_cw'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_ccw'),
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_ccw')
    ]

    common_kwargs_of_streams = {
        "downsample_factor": None,
        "filter_disable": False
    }
    expected_tags_of_streams = [
        'wiregrid, wg_time_constant, wg_inserting, hwp_cw',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_cw',
        'wiregrid, wg_time_constant, hwp_change_cw_to_stop',
        'wiregrid, wg_time_constant, hwp_change_stop_to_ccw',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_ccw',
        'wiregrid, wg_time_constant, wg_ejecting, hwp_ccw'
    ]
    expected_calls_of_streams = [
        call(tag=stream_tag, subtype='cal', kwargs=common_kwargs_of_streams)
        for stream_tag in expected_tags_of_streams
    ]

    for client in wiregrid.run.CLIENTS['smurf']:
        assert client.take_bias_steps.start.call_args_list == expected_calls_of_bias_steps
        assert client.stream.start.call_args_list == expected_calls_of_streams
        assert client.stream.stop.call_count == 6

    assert wiregrid.run.wiregrid.rotate.call_count == 3


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_ccw_el90():
    # Setup all mock clients
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 90, 0)
    wiregrid.run.hwp._get_direction = MagicMock(return_value='ccw')  # ccw
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='outside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid.run.wiregrid.rotate = MagicMock()

    wiregrid.time_constant(num_repeats=1)

    # just make sure bias_steps and streams because other functions are already
    # tested separately.
    expected_calls_of_bias_steps = [
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_ccw, wg_el90'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_ccw, wg_el90'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_cw, wg_el90'),
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_cw, wg_el90')
    ]

    common_kwargs_of_streams = {
        "downsample_factor": None,
        "filter_disable": False
    }
    expected_tags_of_streams = [
        'wiregrid, wg_time_constant, wg_inserting, hwp_ccw, wg_el90',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_ccw, wg_el90',
        'wiregrid, wg_time_constant, hwp_change_ccw_to_stop, wg_el90',
        'wiregrid, wg_time_constant, hwp_change_stop_to_cw, wg_el90',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_cw, wg_el90',
        'wiregrid, wg_time_constant, wg_ejecting, hwp_cw, wg_el90'
    ]
    expected_calls_of_streams = [
        call(tag=stream_tag, subtype='cal', kwargs=common_kwargs_of_streams)
        for stream_tag in expected_tags_of_streams
    ]

    for client in wiregrid.run.CLIENTS['smurf']:
        assert client.take_bias_steps.start.call_args_list == expected_calls_of_bias_steps
        assert client.stream.start.call_args_list == expected_calls_of_streams
        assert client.stream.stop.call_count == 6

    assert wiregrid.run.wiregrid.rotate.call_count == 3


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_repeats():
    # Setup all mock clients
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 50, 0)
    wiregrid.run.hwp._get_direction = MagicMock(return_value='cw')  # cw
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='outside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid.run.wiregrid.rotate = MagicMock()

    wiregrid.time_constant(num_repeats=2)

    # just make sure bias_steps and streams because other functions are already
    # tested separately.
    expected_calls_of_bias_steps = [
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_cw'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_cw'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_ccw'),
        call(tag='wiregrid, wg_time_constant, wg_inserted, hwp_cw'),
        call(tag='wiregrid, wg_time_constant, wg_ejected, hwp_cw')
    ]

    common_kwargs_of_streams = {
        "downsample_factor": None,
        "filter_disable": False
    }
    expected_tags_of_streams = [
        'wiregrid, wg_time_constant, wg_inserting, hwp_cw',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_cw',
        'wiregrid, wg_time_constant, hwp_change_cw_to_stop',
        'wiregrid, wg_time_constant, hwp_change_stop_to_ccw',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_ccw',
        'wiregrid, wg_time_constant, hwp_change_ccw_to_stop',
        'wiregrid, wg_time_constant, hwp_change_stop_to_cw',
        'wiregrid, wg_time_constant, wg_stepwise, hwp_cw',
        'wiregrid, wg_time_constant, wg_ejecting, hwp_cw'
    ]
    expected_calls_of_streams = [
        call(tag=stream_tag, subtype='cal', kwargs=common_kwargs_of_streams)
        for stream_tag in expected_tags_of_streams
    ]

    for client in wiregrid.run.CLIENTS['smurf']:
        assert client.take_bias_steps.start.call_args_list == expected_calls_of_bias_steps
        assert client.stream.start.call_args_list == expected_calls_of_streams
        assert client.stream.stop.call_count == 9

    assert wiregrid.run.wiregrid.rotate.call_count == 4


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_num_repeats_failed():
    with pytest.raises(RuntimeError):
        wiregrid.time_constant(num_repeats=-2)


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_wiregrid_position_failed():
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 50, 0)
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='inside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()

    with pytest.raises(RuntimeError):
        wiregrid.time_constant(num_repeats=1)


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_time_constant_hwp_direction_failed():
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 50, 0)
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = \
        create_actuator_client(motor=1, position='outside')
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()

    # Set up expected raise from invalid direction
    wiregrid.run.hwp._get_direction = MagicMock(return_value=None)
    wiregrid.run.hwp._get_direction.side_effect = RuntimeError("The HWP direction is unknown. Aborting...")

    with pytest.raises(RuntimeError):
        wiregrid.time_constant(num_repeats=1)



"""This module is responsible for commanding all active pysmurf-controller
agents.

If a SMuRF fails to perform an operation it will be dropped from the global
``CLIENTS`` list, removing it from further use during that schedule. If the
number of active SMuRFs falls below the configured ``smurf_failure_threshold``
an exception will be raised, halting observations.

:func:`initialize() <sorunlib.__init__.initialize>` must be run to re-add
previously failed SMuRFs to the ``CLIENTS`` list.

"""

import time

from ocs.client_http import ControlClientError

import sorunlib as run
from sorunlib._internal import check_response, check_started

# Timing between commanding separate SMuRF Controllers
# Yet to be determined in the field. Eventually might need this to be unique
# per operation. Also, move to configuration file once sorunlib has one.
CRYO_WAIT = 120


def _wait_for_cryo(time_):
    if time_ is None:
        wait = CRYO_WAIT
    else:
        wait = time_
    time.sleep(wait)


def _check_smurf_threshold():
    cfg = run.config.load_config()
    threshold = cfg['smurf_failure_threshold']
    remaining = len(run.CLIENTS['smurf'])
    if remaining < threshold:
        error = 'Functional SMuRF count below failure threshold ' + \
                f'({remaining} < {threshold}).'
        raise RuntimeError(error)


def _run_op(operation, concurrent, settling_time, **kwargs):
    """Run operation across all active SMuRF controllers.

    Additional kwargs are passed directly to the ``.start()`` call.

    Args:
        operation (str): Operation name.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    clients_to_remove = []

    # Start operation
    for smurf in run.CLIENTS['smurf']:
        op = smurf.__getattribute__(operation)
        op.start(**kwargs)
        # Run in series
        if not concurrent:
            resp = op.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to perform {operation} on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

            # Allow cryo to settle
            _wait_for_cryo(settling_time)

    # Run in parallel
    if concurrent:
        for smurf in run.CLIENTS['smurf']:
            op = smurf.__getattribute__(operation)
            resp = op.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to perform {operation} on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

    # Remove failed SMuRF clients
    for client in clients_to_remove:
        run.CLIENTS['smurf'].remove(client)

    # Check if enough SMuRFs remain
    _check_smurf_threshold()


def set_targets(targets):
    """Set the target pysmurf-controller Agents that sorunlib will command.

    Args:
        targets (list): List of pysmurf-controller instance-ids to send future
            commands to. This must be a subset of the currently active clients
            list.

    Notes:
        This modifies the global ``sorunlib.CLIENTS`` list.

    """
    _smurf_clients = []
    for smurf in run.CLIENTS['smurf']:
        if smurf.instance_id in targets:
            _smurf_clients.append(smurf)

    run.CLIENTS['smurf'] = _smurf_clients


def bias_step(tag=None, concurrent=True, settling_time=None):
    """Perform a bias step on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    _run_op('take_bias_steps',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def iv_curve(tag=None, iv_kwargs=None, concurrent=True, settling_time=None):
    """Perform an iv curve on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        iv_kwargs (dict, optional): Additional keyword arguments to pass to
            ``take_iv``.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    _run_op('take_iv',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag,
            kwargs=iv_kwargs)


def uxm_setup(concurrent=True, settling_time=0):
    """Perform first-time setup procedure for a UXM.

    Args:
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('uxm_setup',
            concurrent=concurrent,
            settling_time=settling_time)


def uxm_relock(test_mode=False, concurrent=True, settling_time=0):
    """Relocks detectors to existing tune if setup has already been run.

    Args:
        test_mode (bool): Run uxm_setup() task in test_mode, removing emulated
            wait times.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    if test_mode:
        _run_op('uxm_relock',
                concurrent=concurrent,
                settling_time=settling_time,
                test_mode=test_mode)
    else:
        _run_op('uxm_relock',
                concurrent=concurrent,
                settling_time=settling_time)


def bias_dets(rfrac=None, concurrent=True, settling_time=0):
    """Bias the detectors on all SMuRF Controllers.

    Args:
        rfrac (float, tuple, optional): Target rfrac to aim for. If a float
            is provided, bias voltages will be chosen such that the median
            rfrac of each bias group is as close as possible to this value. If
            a tuple is provided, bias voltages will be chosen to maximize the
            number of channels within the given range. If None, the agent
            defaults are used. Defaults to None.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('bias_dets',
            concurrent=concurrent,
            settling_time=settling_time,
            rfrac=rfrac)


def set_biases(bias, bias_group=None, concurrent=True, settling_time=0):
    """Set the detector biases on all SMuRF Controllers.

    Args:
        bias: (int, float, list)
            Biases to set. If a float is passed, this will be used for all
            specified bias groups. If a list of floats is passed, it must be
            the same size of the list of bias groups.
        bias_group: (int, list, optional)
            Bias group or list of bias groups to set. If None, will set all
            bias groups. Defaults to None.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controlls if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('set_biases',
            concurrent=concurrent,
            settling_time=settling_time,
            bias=bias,
            bgs=bias_group)


def zero_biases(bias_group=None, concurrent=True, settling_time=0):
    """Set the detector biases on all SMuRF Controllers.

    Args:
        bias_group: (int, list, optional)
            Bias group or list of bias groups to set. If None, will zero all
            bias groups. Defaults to None.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controlls if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('zero_biases',
            concurrent=concurrent,
            settling_time=settling_time,
            bgs=bias_group)


def take_bgmap(tag=None, concurrent=True, settling_time=0):
    """Take a bgmap on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('take_bgmap',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def take_noise(tag=None, concurrent=True, settling_time=0):
    """Measure noise statistics from a short, 30 second, timestream.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('take_noise',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def shutdown(concurrent=True, settling_time=0):
    """Safely shutdown the SMuRFs.

    This turns off tones, flux-ramp voltage and amplifier biases on all SMuRF
    Controllers.

    Args:
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('all_off',
            concurrent=concurrent,
            settling_time=settling_time)


def stream(state, tag=None, subtype=None, downsample_factor=None, filter_disable=False):
    """Stream data on all SMuRF Controllers.

    Args:
        state (str): Streaming state, either 'on' or 'off'.
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        subtype (str, optional): Operation subtype used to tag the stream.
        downsample_factor (int, optional): Downsample factor. If None, this will be
            pulled from the device cfg.
        filter_disable (bool, optional): If true, will disable the downsample filter
            before streaming.

    """
    clients_to_remove = []
    kwargs = {'downsample_factor': downsample_factor,
              'filter_disable': filter_disable}

    if state.lower() == 'on':
        for smurf in run.CLIENTS['smurf']:
            smurf.stream.start(subtype=subtype, tag=tag, kwargs=kwargs)

        for smurf in run.CLIENTS['smurf']:
            resp = smurf.stream.status()
            try:
                check_started(smurf, resp, timeout=120)
            except RuntimeError as e:
                print(f"Failed to start stream on {smurf}, removing from targets list.")
                print(e)
                smurf.stream.stop()
                clients_to_remove.append(smurf)

    else:
        print('Stopping SMuRF streams.')
        for smurf in run.CLIENTS['smurf']:
            try:
                smurf.stream.stop()
            # Handles case where agent becomes unreachable
            except ControlClientError as e:
                print(f"Failed to stop stream on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

        # Remove failed SMuRF clients
        for client in clients_to_remove:
            run.CLIENTS['smurf'].remove(client)
        clients_to_remove = []

        for smurf in run.CLIENTS['smurf']:
            print(f'Waiting for stream from {smurf.instance_id} to stop.')
            resp = smurf.stream.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to stop stream on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

    # Remove failed SMuRF clients
    for client in clients_to_remove:
        run.CLIENTS['smurf'].remove(client)

    # Check if enough SMuRFs remain
    _check_smurf_threshold()
